from sgx import SgxClient
from sgx.sgx_rpc_handler import SgxServerError
from sgx.sgx_zmq import SgxZmqServerError
import os
from time import sleep
from dotenv import load_dotenv
import random
import coincurve
import binascii
import secrets
import hashlib

load_dotenv()


def bxor(b1, b2):
    parts = []
    for b1, b2 in zip(b1, b2):
        parts.append(bytes([b1 ^ b2]))
    return b''.join(parts)


def decrypt(ciphertext, secret_key):
    xor_val = bxor(ciphertext, secret_key)
    ret_val = binascii.hexlify(xor_val)
    return str(int(ret_val.decode(), 16))


def convert_g2_point_to_hex(data):
    data_hexed = ""
    for coord in data:
        temp = hex(int(coord))[2:]
        while (len(temp) < 64):
            temp = '0' + temp
        data_hexed += temp
    return data_hexed


def perform_complaint(sgx, t, poly_name, public_key, corrupted_secret_key_contribution):
    response = sgx.complaint_response(poly_name, 1)
    share, dh_key = response.share, response.dh_key

    ecdh_key, _ = (coincurve.PublicKey(bytes.fromhex("04" + public_key[2:])).multiply(
                coincurve.keys.PrivateKey.from_hex(dh_key).secret)).point()
    ecdh_key = hex(ecdh_key)[2:]
    while len(ecdh_key) < 64:
        ecdh_key = '0' + ecdh_key

    derived_key = hashlib.sha256(bytes.fromhex(ecdh_key)).digest()

    decrypted_key = decrypt(bytes.fromhex(corrupted_secret_key_contribution), derived_key)

    mult_g2 = sgx.mult_g2(decrypted_key)
    share = share.split(':')
    assert share == mult_g2

    verification_vector_mult = response.verification_vector_mult
    assert len(verification_vector_mult) == t


def perform_dkg(t, n, with_0x=True, with_v2=True, with_complaint=False, with_zmq=False):
    sgx = None
    if not with_zmq:
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'), n=n, t=t)
    else:
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'),
                        n=n, t=t, zmq=True).zmq

    public_keys = []
    key_name = []

    random_dkg_id = random.randint(0, 10**50)

    for i in range(n):
        generated_key = sgx.generate_key()
        if with_0x:
            public_keys.append(generated_key.public_key)
        else:
            public_keys.append(generated_key.public_key[2:])
        key_name.append(generated_key.name)
        sleep(1)

    for i in range(n):
        poly_name = (
            "POLY:SCHAIN_ID:"
            f"{str(0)}"
            ":NODE_ID:"
            f"{str(i)}"
            ":DKG_ID:"
            f"{str(random_dkg_id)}"
        )
        from sgx.sgx_rpc_handler import DkgPolyStatus
        response = sgx.generate_dkg_poly(poly_name)
        if response == DkgPolyStatus.FAIL:
            raise TypeError("failed generate dkg poly for " + str(i))
        sleep(5)

    verification_vector = []
    for i in range(n):
        poly_name = (
            "POLY:SCHAIN_ID:"
            f"{str(0)}"
            ":NODE_ID:"
            f"{str(i)}"
            ":DKG_ID:"
            f"{str(random_dkg_id)}"
        )
        verification_vector.append(sgx.get_verification_vector(poly_name))
        sleep(5)

    hexed_vv = []

    for vv in verification_vector:
        cur_hexed = ""
        for elem in vv:
            cur_hexed += convert_g2_point_to_hex(elem)
        hexed_vv.append(cur_hexed)

    secret_key_contribution = []
    for i in range(n):
        poly_name = (
            "POLY:SCHAIN_ID:"
            f"{str(0)}"
            ":NODE_ID:"
            f"{str(i)}"
            ":DKG_ID:"
            f"{str(random_dkg_id)}"
        )

        if with_v2:
            secret_key_contribution.append(
                sgx.get_secret_key_contribution_v2(poly_name, public_keys))
        else:
            print("KEYS", public_keys)
            secret_key_contribution.append(
                sgx.get_secret_key_contribution(poly_name, public_keys))
        sleep(5)

    if not with_complaint:
        for i in range(n):
            for j in range(n):
                if with_v2:
                    if not sgx.verify_secret_share_v2(
                            hexed_vv[j],
                            key_name[i],
                            secret_key_contribution[j][192*i:192*(i + 1)], i):
                        raise ValueError(f'{i} failed to verify {j}')
                    sleep(5)
                else:
                    if not sgx.verify_secret_share(
                            hexed_vv[j],
                            key_name[i],
                            secret_key_contribution[j][192*i:192*(i + 1)], i):
                        raise ValueError(f'{i} failed to verify {j}')
                    sleep(5)

        public_keys = sgx.calculate_all_bls_public_keys(hexed_vv)

        for i in range(n):
            poly_name = (
                "POLY:SCHAIN_ID:"
                f"{str(0)}"
                ":NODE_ID:"
                f"{str(i)}"
                ":DKG_ID:"
                f"{str(random_dkg_id)}"
            )
            bls_key_name = (
                "BLS_KEY:SCHAIN_ID:"
                f"{str(0)}"
                ":NODE_ID:"
                f"{str(i)}"
                ":DKG_ID:"
                f"{str(random_dkg_id)}"
            )

            if with_v2:
                sgx.create_bls_private_key_v2(
                    poly_name,
                    bls_key_name,
                    key_name[i],
                    "".join(secret_key_contribution[j][192*i:192*(i + 1)] for j in range(n)))
            else:
                sgx.create_bls_private_key(
                    poly_name,
                    bls_key_name,
                    key_name[i],
                    "".join(secret_key_contribution[j][192*i:192*(i + 1)] for j in range(n)))

            public_key = sgx.get_bls_public_key(bls_key_name)

            assert ":".join(public_key) == public_keys[i]

            message = secrets.token_hex(32)

            signature_share = sgx.bls_sign(bls_key_name, message)
            splitted_signature = signature_share.split(':')
            assert len(splitted_signature) == 4

            assert len(splitted_signature[0]) > 0
            assert len(splitted_signature[0]) < 78

            assert len(splitted_signature[1]) > 0
            assert len(splitted_signature[1]) < 78

            assert len(splitted_signature[2]) > 0
            assert len(splitted_signature[2]) < 78

            assert int(splitted_signature[3]) < 1000

            sleep(5)
    else:
        corrupted_secret_key_contribution = secret_key_contribution[0]
        secret_key_contribution[0] = secret_key_contribution[1]

        for i in range(n):
            for j in range(n):
                if j == 0:
                    if with_v2:
                        assert not sgx.verify_secret_share_v2(
                                hexed_vv[j],
                                key_name[i],
                                secret_key_contribution[j][192*i:192*(i + 1)], i)
                    else:
                        assert not sgx.verify_secret_share(
                                hexed_vv[j],
                                key_name[i],
                                secret_key_contribution[j][192*i:192*(i + 1)], i)
                else:
                    if with_v2:
                        assert sgx.verify_secret_share_v2(
                                hexed_vv[j],
                                key_name[i],
                                secret_key_contribution[j][192*i:192*(i + 1)], i)
                    else:
                        assert sgx.verify_secret_share(
                                hexed_vv[j],
                                key_name[i],
                                secret_key_contribution[j][192*i:192*(i + 1)], i)
                sleep(5)

        poly_name = (
                "POLY:SCHAIN_ID:"
                f"{str(0)}"
                ":NODE_ID:"
                f"{str(0)}"
                ":DKG_ID:"
                f"{str(random_dkg_id)}"
            )
        perform_complaint(
                        sgx,
                        t,
                        poly_name,
                        public_keys[1],
                        corrupted_secret_key_contribution[192:256]
                        )


def perform_poly_existence(with_zmq=False):
    sgx = None
    if not with_zmq:
        print("TESTING SGX WITHOUT ZMQ")
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'), n=2, t=2)
    else:
        print("TESTING SGX WITH ZMQ")
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'),
                        n=2, t=2, zmq=True).zmq

    random_dkg_id = random.randint(0, 10**50)

    poly_name = (
            "POLY:SCHAIN_ID:"
            f"{str(0)}"
            ":NODE_ID:"
            f"{str(0)}"
            ":DKG_ID:"
            f"{str(random_dkg_id)}"
        )
    if with_zmq:
        from sgx.sgx_zmq import DkgPolyStatus
        assert sgx.generate_dkg_poly(poly_name) == DkgPolyStatus.NEW_GENERATED
    else:
        from sgx.sgx_rpc_handler import DkgPolyStatus
        assert sgx.generate_dkg_poly(poly_name) == DkgPolyStatus.NEW_GENERATED
    assert sgx.is_poly_exists(poly_name)
    poly_name_incorrect = (
            "POLY:SCHAIN_ID:"
            f"{str(0)}"
            ":NODE_ID:"
            f"{str(0)}"
            ":DKG_ID:"
            f"{str(random_dkg_id+1)}"
        )
    assert not sgx.is_poly_exists(poly_name_incorrect)
    response = sgx.generate_dkg_poly(poly_name)
    assert response == DkgPolyStatus.PREEXISTING


def perform_import(with_zmq=False, with_0x=False):
    if not with_zmq:
        print("TESTING SGX WITHOUT ZMQ")
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'), n=2, t=2)
    else:
        print("TESTING SGX WITH ZMQ")
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'),
                        n=2, t=2, zmq=True).zmq

    random_dkg_id = random.randint(0, 10**50)

    bls_key_name = (
                "BLS_KEY:SCHAIN_ID:"
                f"{str(0)}"
                ":NODE_ID:"
                f"{str(0)}"
                ":DKG_ID:"
                f"{str(random_dkg_id)}"
    )

    insecure_bls_private_key = "f253bad7b1f62b8ff60bbf451cf2e8e9ebb5d6e9bff450c55b8d5504b8c63d3"

    response = sgx.import_bls_private_key(bls_key_name, insecure_bls_private_key)

    assert len(response) > 0

    message = secrets.token_hex(32)
    if with_0x:
        message = "0x" + message

    signature_share = sgx.bls_sign(bls_key_name, message)
    splitted_signature = signature_share.split(':')
    assert len(splitted_signature) == 4

    assert len(splitted_signature[0]) > 0
    assert len(splitted_signature[0]) < 78

    assert len(splitted_signature[1]) > 0
    assert len(splitted_signature[1]) < 78

    assert len(splitted_signature[2]) > 0
    assert len(splitted_signature[2]) < 78

    assert int(splitted_signature[3]) < 1000


def perform_delete(with_zmq=False):
    if not with_zmq:
        print("TESTING SGX WITHOUT ZMQ")
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'), n=2, t=2)
    else:
        print("TESTING SGX WITH ZMQ")
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'),
                        n=2, t=2, zmq=True).zmq

    random_dkg_id = random.randint(0, 10**50)

    bls_key_name = (
                "BLS_KEY:SCHAIN_ID:"
                f"{str(0)}"
                ":NODE_ID:"
                f"{str(0)}"
                ":DKG_ID:"
                f"{str(random_dkg_id)}"
    )

    insecure_bls_private_key = "f253bad7b1f62b8ff60bbf451cf2e8e9ebb5d6e9bff450c55b8d5504b8c63d3"

    response = sgx.import_bls_private_key(bls_key_name, insecure_bls_private_key)

    assert len(response) > 0

    sgx.delete_bls_key(bls_key_name)
    try:
        sgx.delete_bls_key(bls_key_name)
    except SgxServerError as e:
        str_error = f'deleteBlsKeyImpl failed:deleteBlsKeyImpl:BLS key not found: {bls_key_name}'
        assert str(e) == str_error
    except SgxZmqServerError as e:
        str_error = f'deleteBlsKeyImpl failed:deleteBlsKeyImpl:BLS key not found: {bls_key_name}'
        assert str(e) == str_error


def perform_helper(with_zmq=False):
    sgx = None
    if not with_zmq:
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'))
    else:
        sgx = SgxClient(os.environ['SERVER'], path_to_cert=os.environ.get('CERT_PATH'),
                        zmq=True).zmq
    assert sgx.get_server_status() == 0
    assert isinstance(sgx.get_server_version(), str)


def test_dkg():
    perform_dkg(2, 2, with_0x=True)
    print("TEST WITH 0x PREFIX PASSED")
    perform_dkg(2, 2, with_0x=False)
    print("TEST WITHOUT 0x PREFIX PASSED")


def test_old_dkg():
    perform_dkg(2, 2, with_0x=True, with_v2=False)
    print("TEST OLD DKG WITH 0x PREFIX PASSED")
    perform_dkg(2, 2, with_0x=False, with_v2=False)
    print("TEST OLD DKG WITHOUT 0x PREFIX PASSED")


def test_dkg_zmq():
    perform_dkg(2, 2, with_0x=True, with_v2=False, with_zmq=True)
    print("TEST DKG WITH ZMQ PASSED")


def test_dkg_complaint():
    perform_dkg(2, 2, with_complaint=True)
    perform_dkg(2, 2, with_v2=False, with_complaint=True, with_zmq=True)
    print("TEST DKG COMPLAINT PASSED")


def test_poly_existence():
    perform_poly_existence()
    perform_poly_existence(with_zmq=True)
    print("TEST POLY EXISTENCE PASSED")


def test_import():
    perform_import()
    perform_import(with_0x=True)
    perform_import(with_zmq=True)
    perform_import(with_zmq=True, with_0x=True)
    print("TEST IMPORT BLS KEY PASSED")


def test_delete():
    perform_delete()
    perform_delete(with_zmq=True)
    print("TEST DELETE BLS KEY PASSED")


def test_helper_functions():
    perform_helper()
    perform_helper(with_zmq=True)
    print("TEST HELPER FUNCTIONS PASSED")



import os

CUR_DIR = os.path.dirname(os.path.realpath(__file__))
GENERATE_SCRIPT_PATH = os.path.join(CUR_DIR, 'generate.sh')

DEFAULT_TIMEOUT = 10
SGX_RESPONSE_TIMEOUT = 60  # seconds
SGX_ZMQ_RESPONSE_TIMEOUT_MS = 5000

CSR_FILENAME = 'sgx.csr'
KEY_FILENAME = 'sgx.key'
CRT_FILENAME = 'sgx.crt'



#    -*- coding: utf-8 -*-
#
#     This file is part of sgx.py
#
#     Copyright (C) 2021 SKALE Labs
#
#     sgx.py is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     sgx.py is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with sgx.py.  If not, see <https://www.gnu.org/licenses/>.

import binascii
import json
import logging
import os
from dataclasses import dataclass
from enum import Enum
from time import sleep
from sgx.utils import public_key_to_address

from eth_utils.conversions import add_0x_prefix, remove_0x_prefix
import pem
import zmq
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, rsa, padding
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from urllib.parse import urlparse

from sgx.constants import DEFAULT_TIMEOUT, SGX_ZMQ_RESPONSE_TIMEOUT_MS, CRT_FILENAME, KEY_FILENAME
from sgx.utils import SgxError


logger = logging.getLogger(__name__)


ZMQ_PORT = 1031
MAX_RETRIES = 10


class SgxZmqServerError(SgxError):
    pass


class SgxZmqUnreachableError(SgxError):
    pass


class DkgPolyStatus(Enum):
    FAIL = 0
    NEW_GENERATED = 1
    PREEXISTING = 2


@dataclass
class Account:
    name: str
    address: str
    public_key: str


@dataclass
class ComplaintResponse:
    share: str
    dh_key: str
    verification_vector_mult: str


class SgxZmq:
    def __init__(self, sgx_endpoint, path_to_cert=None, n=None, t=None):
        self.sgx_endpoint = get_provider(sgx_endpoint)
        self.path_to_cert = path_to_cert
        if n:
            self.n = n
        if t:
            self.t = t
        self.cert = self.__read_cert()
        self.ctx = zmq.Context()
        self.__init_method_types()
        self.sockets = dict()

    def __del__(self):
        for socket in self.sockets.values():
            socket.close()
        self.ctx.destroy()

    def ecdsa_sign(self, key_name, transaction_hash):
        params = dict()
        params['base'] = 10
        params['keyName'] = key_name
        params['messageHash'] = transaction_hash
        response = self.__send_request('ecdsaSignMessageHash', params)
        signature = response['result']
        vrs = (signature['signature_v'], signature['signature_r'], signature['signature_s'])
        return vrs

    def generate_key(self):
        params = dict()
        response = self.__send_request('generateECDSAKey', params)
        key_name = response['keyName']
        public_key = response['publicKey']
        public_key = add_0x_prefix(public_key)
        address = public_key_to_address(public_key)
        return Account(name=key_name, address=address, public_key=public_key)

    def get_public_key(self, keyName):
        params = dict()
        params['keyName'] = keyName
        response = self.__send_request('getPublicECDSAKey', params)
        publicKey = response['publicKey']
        return publicKey

    def generate_dkg_poly(self, poly_name):
        if self.is_poly_exists(poly_name):
            return DkgPolyStatus.PREEXISTING
        params = dict()
        params['polyName'] = poly_name
        params['t'] = self.t
        response = self.__send_request('generateDKGPoly', params)
        if response['status'] == 0:
            return DkgPolyStatus.NEW_GENERATED
        else:
            return DkgPolyStatus.FAIL

    def get_verification_vector(self, poly_name):
        params = dict()
        params['polyName'] = poly_name
        params['n'] = self.n
        params['t'] = self.t
        response = self.__send_request('getVerificationVector', params)
        verification_vector = response['verificationVector']
        return verification_vector

    def get_secret_key_contribution(self, poly_name, public_keys):
        public_keys = list(map(remove_0x_prefix, public_keys))
        params = dict()
        params['polyName'] = poly_name
        params['n'] = self.n
        params['t'] = self.t
        params['publicKeys'] = public_keys
        response = self.__send_request('getSecretShare', params)
        secret_key_contribution = response['secretShare']
        return secret_key_contribution

    def get_server_status(self):
        response = self.__send_request('getServerStatus')
        return response['status']

    def get_server_version(self):
        response = self.__send_request('getServerVersion')
        return response['version']

    def verify_secret_share(self, public_shares, eth_key_name, secret_share, index):
        params = dict()
        params['publicShares'] = public_shares
        params['ethKeyName'] = eth_key_name
        params['secretShare'] = secret_share
        params['n'] = self.n
        params['t'] = self.t
        params['index'] = index
        response = self.__send_request('dkgVerification', params)
        result = response['result']
        return result

    def create_bls_private_key(self, poly_name, bls_key_name, eth_key_name, secret_shares):
        params = dict()
        params['polyName'] = poly_name
        params['blsKeyName'] = bls_key_name
        params['ethKeyName'] = eth_key_name
        params['secretShare'] = secret_shares
        params['n'] = self.n
        params['t'] = self.t
        response = self.__send_request('createBLSPrivateKey', params)
        return response['status'] == 0

    def get_bls_public_key(self, bls_key_name):
        params = dict()
        params['blsKeyName'] = bls_key_name
        response = self.__send_request('getBLSPublicKeyShare', params)
        return response['blsPublicKeyShare']

    def complaint_response(self, poly_name, idx):
        params = dict()
        params['polyName'] = poly_name
        params['n'] = self.n
        params['t'] = self.t
        params['ind'] = idx
        response = self.__send_request('complaintResponse', params)
        return ComplaintResponse(
            share=response['share*G2'],
            dh_key=response['dhKey'],
            verification_vector_mult=response['verificationVectorMult'],
        )

    def mult_g2(self, to_mult):
        params = dict()
        params['x'] = to_mult
        response = self.__send_request('multG2', params)
        return response['x*G2']

    def import_bls_private_key(self, key_share_name, key_share):
        params = dict()
        params['keyShareName'] = key_share_name
        params['keyShare'] = key_share
        response = self.__send_request('importBLSKeyShare', params)
        encrypted_key = response['encryptedKeyShare']
        return encrypted_key

    def is_poly_exists(self, poly_name):
        params = dict()
        params['polyName'] = poly_name
        response = self.__send_request('isPolyExists', params)
        is_exists = response['IsExist']
        return is_exists

    def delete_bls_key(self, bls_key_name):
        params = dict()
        params['blsKeyName'] = bls_key_name
        response = self.__send_request('deleteBlsKey', params)
        result = response['deleted']

        return result

    def calculate_all_bls_public_keys(self, verification_vectors):
        params = dict()
        params['t'] = self.t
        params['n'] = self.n
        params['publicShares'] = verification_vectors
        response = self.__send_request('calculateAllBLSPublicKeys', params)
        result = response['publicKeys']

        return result

    def bls_sign(self, bls_key_name, message_hash):
        params = dict()
        params['keyShareName'] = bls_key_name
        params['messageHash'] = message_hash
        params['t'] = self.t
        params['n'] = self.n
        response = self.__send_request('blsSignMessageHash', params)
        result = response['signatureShare']

        return result

    def __send_request(self, method, params=None):
        if not params:
            params = dict()
        params['type'] = self.method_to_type[method]
        if self.path_to_cert:
            params['cert'] = self.cert
            msgSig = self.__sign_msg(params)
            params['msgSig'] = msgSig
        msg = json.dumps(params, separators=(',', ':'))
        p_id = os.getpid()
        if not self.sockets.get(p_id):
            socket_p_id = self.ctx.socket(zmq.DEALER)
            socket_p_id.setsockopt_string(zmq.IDENTITY, '135:14603077656239261618')
            socket_p_id.setsockopt(zmq.LINGER, 0)
            socket_p_id.setsockopt(zmq.SNDTIMEO, SGX_ZMQ_RESPONSE_TIMEOUT_MS)
            socket_p_id.setsockopt(zmq.RCVTIMEO, SGX_ZMQ_RESPONSE_TIMEOUT_MS)
            socket_p_id.connect(self.sgx_endpoint)
            self.sockets[p_id] = socket_p_id
        socket = self.sockets[p_id]
        socket.send_string(msg)
        # await reply
        response_str = None
        for _ in range(MAX_RETRIES):
            try:
                response_str = socket.recv().decode()
            except zmq.ZMQError:
                pass
            if response_str:
                break
            sleep(DEFAULT_TIMEOUT)
        if not response_str:
            raise SgxZmqUnreachableError('Max retries exceeded for sgx connection')
        response = json.loads(response_str)
        if (
            response.get('errorMessage') is not None and len(response.get('errorMessage'))
        ) or response['status']:
            raise SgxZmqServerError(response['errorMessage'])
        return response

    def __sign_msg(self, to_sign):
        msg = json.dumps(to_sign, separators=(',', ':'))
        msg = msg.replace(' ', '')
        key_path = os.path.join(self.path_to_cert, KEY_FILENAME)
        with open(key_path, 'rb') as key_file:
            private_key_bytes = key_file.read()
        private_key = load_pem_private_key(private_key_bytes, password=None)
        data = msg.encode()
        if isinstance(private_key, ec.EllipticCurvePrivateKey):
            signature = private_key.sign(data, ec.ECDSA(hashes.SHA256()))
        elif isinstance(private_key, rsa.RSAPrivateKey):
            signature = private_key.sign(data, padding.PKCS1v15(), hashes.SHA256())
        else:
            raise TypeError('Unsupported private key type for signing')
        return binascii.hexlify(signature).decode()

    def __read_cert(self):
        crt_path = os.path.join(self.path_to_cert, CRT_FILENAME)
        crt = pem.parse_file(crt_path)
        return str(crt[0])

    def __init_method_types(self):
        self.method_to_type = dict()
        self.method_to_type['ecdsaSignMessageHash'] = 'ECDSASignReq'
        self.method_to_type['generateECDSAKey'] = 'generateECDSAReq'
        self.method_to_type['getPublicECDSAKey'] = 'getPublicECDSAReq'
        self.method_to_type['generateDKGPoly'] = 'generateDKGPolyReq'
        self.method_to_type['getVerificationVector'] = 'getVerificationVectorReq'
        self.method_to_type['getSecretShare'] = 'getSecretShareReq'
        self.method_to_type['getServerStatus'] = 'getServerStatusReq'
        self.method_to_type['getServerVersion'] = 'getServerVersionReq'
        self.method_to_type['dkgVerification'] = 'dkgVerificationReq'
        self.method_to_type['createBLSPrivateKey'] = 'createBLSPrivateReq'
        self.method_to_type['getBLSPublicKeyShare'] = 'getBLSPublicReq'
        self.method_to_type['complaintResponse'] = 'complaintResponseReq'
        self.method_to_type['importBLSKeyShare'] = 'importBLSReq'
        self.method_to_type['isPolyExists'] = 'isPolyExistsReq'
        self.method_to_type['deleteBlsKey'] = 'deleteBLSKeyReq'
        self.method_to_type['calculateAllBLSPublicKeys'] = 'getAllBLSPublicReq'
        self.method_to_type['blsSignMessageHash'] = 'BLSSignReq'
        self.method_to_type['multG2'] = 'multG2Req'


def get_provider(endpoint):
    parsed_endpoint = urlparse(endpoint)
    ip, _ = parsed_endpoint.netloc.split(':')
    zmq_endpoint = 'tcp://' + ip + ':' + str(ZMQ_PORT)
    return zmq_endpoint



#    -*- coding: utf-8 -*-
#
#     This file is part of sgx.py
#
#     Copyright (C) 2019 SKALE Labs
#
#     sgx.py is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     sgx.py is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with sgx.py.  If not, see <https://www.gnu.org/licenses/>.


import copy
import logging
import os
import subprocess
from subprocess import PIPE

from web3 import Web3


logger = logging.getLogger(__name__)


class SgxError(Exception):
    pass


def crop_json(json_data, crop_len=50):
    for key, value in json_data.items():
        if isinstance(value, dict):
            crop_json(value)
        else:
            if isinstance(value, str) and len(value) > crop_len:
                json_data[key] = value[:crop_len] + '...'


def print_request_log(request):
    cropped_request = copy.deepcopy(request)
    crop_json(cropped_request)
    logger.info(f'Send request: {request}')


def print_response_log(response):
    cropped_response = copy.deepcopy(response)
    crop_json(cropped_response)
    logger.info(f'Response received: {cropped_response}')


def run_cmd(cmd, env={}, shell=False):
    logger.info(f'Running: {cmd}')
    res = subprocess.run(
        cmd, shell=shell, stdout=PIPE, stderr=PIPE,
        env={**env, **os.environ}
    )
    if res.returncode:
        logger.error('Error during shell execution:')
        logger.error(res.stderr.decode('UTF-8').rstrip())
        raise subprocess.CalledProcessError(res.returncode, cmd)
    return res


def public_key_to_address(pk):
    hash_ = Web3.keccak(hexstr=str(pk))
    return Web3.to_checksum_address(Web3.to_hex(hash_[-20:]))



#    -*- coding: utf-8 -*-
#
#     This file is part of sgx.py
#
#     Copyright (C) 2019 SKALE Labs
#
#     sgx.py is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as published
#     by the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     sgx.py is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.
#
#     You should have received a copy of the GNU Affero General Public License
#     along with sgx.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
from dataclasses import dataclass
from collections.abc import Mapping
from hexbytes import HexBytes
from eth_account.datastructures import SignedTransaction, SignedMessage
from eth_utils.curried import keccak
from cytoolz import dissoc
from sgx.sgx_rpc_handler import SgxRPCHandler
from sgx.sgx_zmq import SgxZmq
from sgx.utils import public_key_to_address
from eth_account._utils import signing
from eth_account._utils.legacy_transactions import encode_transaction
from eth_account.typed_transactions.typed_transaction import TypedTransaction
from eth_utils.encoding import big_endian_to_int
from eth_utils.conversions import add_0x_prefix, remove_0x_prefix

logger = logging.getLogger(__name__)


@dataclass
class Account:
    name: str
    address: str
    public_key: str


@dataclass
class ComplaintResponse:
    share: str
    dh_key: str
    verification_vector_mult: str


class SgxClient:
    def __init__(self, sgx_endpoint, path_to_cert=None, n=None, t=None, zmq=False):
        self.sgx_endpoint = sgx_endpoint
        self.sgx_rpc_server = SgxRPCHandler(sgx_endpoint, path_to_cert)
        if zmq:
            self.zmq = SgxZmq(sgx_endpoint, path_to_cert, n, t)
        if not path_to_cert:
            logger.warning('Using SgxClient without certificates')
        if n:
            self.n = n
        if t:
            self.t = t

    def generate_key(self):
        key_name, public_key = self.sgx_rpc_server.generate_key()
        public_key = add_0x_prefix(public_key)
        address = public_key_to_address(public_key)
        return Account(name=key_name, address=address, public_key=public_key)

    def get_account(self, key_name):
        key = self.sgx_rpc_server.get_public_key(key_name)
        key = add_0x_prefix(key)
        address = public_key_to_address(key)
        return Account(name=key_name, address=address, public_key=key)

    def sign(self, transaction_dict, key_name):
        if not isinstance(transaction_dict, Mapping):
            raise TypeError('transaction_dict must be dict-like, got %r' % transaction_dict)

        address = self.get_account(key_name).address

        # allow from field, *only* if it matches the private key
        if 'from' in transaction_dict:
            if transaction_dict['from'] == address:
                sanitized_transaction = dissoc(transaction_dict, 'from')
            else:
                raise TypeError(
                    "from field must match key's %s, but it was %s"
                    % (
                        address,
                        transaction_dict['from'],
                    )
                )
        else:
            sanitized_transaction = transaction_dict

        # sign transaction
        (
            v,
            r,
            s,
            rlp_encoded,
        ) = self._sign_transaction_dict(key_name, sanitized_transaction)

        transaction_hash = keccak(rlp_encoded)
        return SignedTransaction(
            raw_transaction=HexBytes(rlp_encoded), hash=HexBytes(transaction_hash), r=r, s=s, v=v
        )

    def sign_hash(self, message, key_name, chain_id):
        msg_hash_bytes = HexBytes(message)
        if len(msg_hash_bytes) != 32:
            raise ValueError('The message hash must be exactly 32-bytes')

        (v, r, s) = self._sign_hash(key_name, msg_hash_bytes, chain_id, eip_1559_identifier=False)
        signature_bytes = signing.to_bytes32(r) + signing.to_bytes32(s) + signing.to_bytes(v)
        return SignedMessage(
            message_hash=msg_hash_bytes, r=r, s=s, v=v, signature=HexBytes(signature_bytes)
        )

    def generate_dkg_poly(self, poly_name):
        return self.sgx_rpc_server.generate_dkg_poly(poly_name, self.t)

    def get_verification_vector(self, poly_name):
        return self.sgx_rpc_server.get_verification_vector(poly_name, self.n, self.t)

    def get_secret_key_contribution(self, poly_name, public_keys):
        public_keys = list(map(remove_0x_prefix, public_keys))
        return self.sgx_rpc_server.get_secret_key_contribution(
            poly_name, public_keys, self.n, self.t
        )

    def get_secret_key_contribution_v2(self, poly_name, public_keys):
        public_keys = list(map(remove_0x_prefix, public_keys))
        return self.sgx_rpc_server.get_secret_key_contribution_v2(
            poly_name, public_keys, self.n, self.t
        )

    def get_server_status(self):
        return self.sgx_rpc_server.get_server_status()

    def get_server_version(self):
        return self.sgx_rpc_server.get_server_version()

    def verify_secret_share(self, public_shares, eth_key_name, secret_share, index):
        return self.sgx_rpc_server.verify_secret_share(
            public_shares, eth_key_name, secret_share, self.n, self.t, index
        )

    def verify_secret_share_v2(self, public_shares, eth_key_name, secret_share, index):
        return self.sgx_rpc_server.verify_secret_share_v2(
            public_shares, eth_key_name, secret_share, self.n, self.t, index
        )

    def create_bls_private_key(self, poly_name, bls_key_name, eth_key_name, secret_shares):
        return self.sgx_rpc_server.create_bls_private_key(
            poly_name, bls_key_name, eth_key_name, secret_shares, self.n, self.t
        )

    def create_bls_private_key_v2(self, poly_name, bls_key_name, eth_key_name, secret_shares):
        return self.sgx_rpc_server.create_bls_private_key_v2(
            poly_name, bls_key_name, eth_key_name, secret_shares, self.n, self.t
        )

    def get_bls_public_key(self, bls_key_name):
        return self.sgx_rpc_server.get_bls_public_key(bls_key_name)

    def complaint_response(self, poly_name, idx):
        share, dh_key, verification_vector_mult = self.sgx_rpc_server.complaint_response(
            poly_name, self.n, self.t, idx
        )
        return ComplaintResponse(
            share=share, dh_key=dh_key, verification_vector_mult=verification_vector_mult
        )

    def mult_g2(self, to_mult):
        return self.sgx_rpc_server.mult_g2(to_mult)

    def import_bls_private_key(self, key_share_name, key_share):
        return self.sgx_rpc_server.import_bls_private_key(key_share_name, key_share)

    def import_ecdsa_private_key(self, key_share_name, key_share):
        return self.sgx_rpc_server.import_ecdsa_private_key(key_share_name, key_share)

    def is_poly_exists(self, poly_name):
        return self.sgx_rpc_server.is_poly_exist(poly_name)

    def delete_bls_key(self, bls_key_name):
        return self.sgx_rpc_server.delete_bls_key(bls_key_name)

    def calculate_all_bls_public_keys(self, verification_vectors):
        return self.sgx_rpc_server.calculate_all_bls_public_keys(
            verification_vectors, self.t, self.n
        )

    def bls_sign(self, bls_key_name, message_hash):
        return self.sgx_rpc_server.bls_sign(bls_key_name, message_hash, self.t, self.n)

    def _sign_transaction_dict(self, eth_key, transaction_dict):
        # generate RLP-serializable transaction, with defaults filled
        unsigned_tx = signing.serializable_unsigned_transaction_from_dict(transaction_dict)

        transaction_hash = unsigned_tx.hash()

        chain_id = transaction_dict['chainId']

        eip_1559_identifier = False

        if isinstance(unsigned_tx, TypedTransaction):
            eip_1559_identifier = True

        (v, r, s) = self._sign_hash(eth_key, transaction_hash, chain_id, eip_1559_identifier)

        # serialize transaction with rlp
        encoded_tx = encode_transaction(unsigned_tx, vrs=(v, r, s))

        return v, r, s, encoded_tx

    def _sign_hash(self, eth_key, transaction_hash, chain_id, eip_1559_identifier=False):
        hash_in_hex = hex(big_endian_to_int(transaction_hash))
        (v_raw, r_raw, s_raw) = self.sgx_rpc_server.ecdsa_sign(eth_key, hash_in_hex)
        if eip_1559_identifier:
            v = int(v_raw)
        else:
            v = signing.to_eth_v(int(v_raw), chain_id)

        r = int(r_raw)
        s = int(s_raw)
        return (v, r, s)



import os
import logging
import secrets
import requests
import json
from time import sleep
from urllib.parse import urlparse

import urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from sgx.constants import (
    GENERATE_SCRIPT_PATH,
    DEFAULT_TIMEOUT,
    CSR_FILENAME,
    CRT_FILENAME,
    KEY_FILENAME,
    SGX_RESPONSE_TIMEOUT
)
from sgx.utils import run_cmd, print_request_log, print_response_log, SgxError


requests.packages.urllib3.disable_warnings(InsecureRequestWarning)  # TODO: Remove
logger = logging.getLogger(__name__)


MAX_RETRIES = 22


class SgxSSLError(SgxError):
    pass


class SgxUnreachableError(SgxError):
    pass


def get_certificate_credentials(crt_dir_path, csr_server):
    key_path = os.path.join(crt_dir_path, KEY_FILENAME)
    crt_path = os.path.join(crt_dir_path, CRT_FILENAME)
    if not os.path.exists(crt_path) or not os.path.exists(key_path):
        csr_path = os.path.join(crt_dir_path, CSR_FILENAME)
        if not os.path.exists(csr_path) or not os.path.exists(key_path):
            generate_csr_credentials(csr_path, key_path)
        csr_hash = sign_certificate(csr_server, csr_path)
        write_crt_to_file(crt_path, csr_server, csr_hash)
    return crt_path, key_path


def generate_csr_credentials(csr_path, key_path):
    certificate_name = secrets.token_hex(nbytes=32)
    run_cmd(["bash", GENERATE_SCRIPT_PATH, csr_path, key_path, certificate_name])


def write_crt_to_file(crt_path, csr_server, csr_hash):
    response = send_request_safe(csr_server, 'getCertificate', {'hash': csr_hash})
    while response['result']['status'] == 1:
        response = send_request_safe(csr_server, 'getCertificate', {'hash': csr_hash})
        sleep(DEFAULT_TIMEOUT)
    crt = response['result']['cert']
    with open(crt_path, "w+") as f:
        f.write(crt)


def sign_certificate(csr_server, csr_path):
    with open(csr_path) as csr_file:
        csr = csr_file.read()
    response = send_request_safe(csr_server, 'signCertificate', {'certificate': csr})
    csr_hash = response['result']['hash']
    return csr_hash


def send_request(url, method, params, path_to_cert=None):
    headers = {'content-type': 'application/json'}
    call_data = {
        "id": 0,
        "jsonrpc": "2.0",
        "method": method,
        "params": params,
    }
    print_request_log(call_data)
    cert = None
    if path_to_cert:
        cert = get_certificate_credentials(
            path_to_cert,
            get_cert_provider(url)
        )

    try:
        response = requests.post(
            url,
            data=json.dumps(call_data),
            headers=headers,
            cert=cert,
            verify=False,
            timeout=SGX_RESPONSE_TIMEOUT
        ).json()
    except requests.exceptions.ConnectionError as err:
        logger.error('Connection to server failed', exc_info=err)
        if isinstance(err.args[0], urllib3.exceptions.MaxRetryError):
            raise SgxUnreachableError('Max retries exceeded for sgx connection')
        raise
    print_response_log(response)
    return response


def send_request_safe(url, method, params=None):
    response = send_request(url, method, params)
    if response.get('error'):
        raise SgxSSLError(response['error']['message'])
    if response['result']['status'] != 0:
        raise SgxSSLError(response['result']['errorMessage'])
    return response


def get_cert_provider(endpoint):
    parsed_endpoint = urlparse(endpoint)
    port = str(parsed_endpoint.port + 1)
    url = 'http://' + parsed_endpoint.hostname + ':' + port
    return url



import urllib
import os
import secrets

import eth_account._utils.legacy_transactions as transactions
import pytest

from eth_account.messages import defunct_hash_message, encode_defunct

from dotenv import load_dotenv
from eth_keys import keys
from hexbytes import HexBytes
import socket
from web3 import Web3


from sgx import SgxClient


load_dotenv()

SGX_URL = os.getenv('SERVER')
GETH_URL = os.getenv('GETH')
ETH_PRIVATE_KEY = os.getenv('ETH_PRIVATE_KEY')

ETH_VALUE_FOR_TESTS = 5 * 10**18


def private_key_to_public(pr):
    pr_bytes = Web3.to_bytes(hexstr=pr)
    pk = keys.PrivateKey(pr_bytes)
    return pk.public_key


def public_key_to_address(pk):
    hash = Web3.keccak(hexstr=str(pk))
    return Web3.to_checksum_address(Web3.to_hex(hash[-20:]))


def private_key_to_address(pr):
    pk = private_key_to_public(pr)
    return public_key_to_address(pk)


@pytest.fixture
def w3():
    return Web3(Web3.HTTPProvider(GETH_URL))


def send_eth_w3(web3, to, value):
    from_address = private_key_to_address(ETH_PRIVATE_KEY)
    nonce = web3.eth.get_transaction_count(from_address)
    tx = {
        'from': from_address,
        'to': to,
        'value': value,
        'gas': 21000,
        'maxFeePerGas': 10**9,
        'maxPriorityFeePerGas': 10,
        'nonce': nonce,
        'chainId': web3.eth.chain_id,
    }
    signed = web3.eth.account.sign_transaction(tx, private_key=ETH_PRIVATE_KEY)
    tx_hash = web3.eth.send_raw_transaction(signed.raw_transaction)
    web3.eth.wait_for_transaction_receipt(tx_hash)


@pytest.fixture
def sgx(w3):
    return SgxClient(SGX_URL, os.getenv('CERT_PATH'))


@pytest.fixture
def account(sgx, w3):
    c = SgxClient(SGX_URL, os.getenv('CERT_PATH'))
    generated_key = c.generate_key()
    key = generated_key.name
    address = c.get_account(key).address
    send_eth_w3(w3, address, ETH_VALUE_FOR_TESTS)
    return key, address


def generate_tx(web3, tx_type=None, no_fee=False):
    txn = {
        'to': os.getenv('TEST_ACCOUNT'),
        'value': 0,
        'gas': 200000,
        'gasPrice': web3.eth.gas_price,
        'chainId': web3.eth.chain_id,
    }
    if tx_type:
        txn['type'] = tx_type
        if tx_type == 2:
            txn.pop('gasPrice', None)
            txn.update({'maxFeePerGas': 10**9, 'maxPriorityFeePerGas': 10})

    return txn


def test_server_connection():
    parsed_url = urllib.parse.urlparse(SGX_URL)
    with socket.create_connection((parsed_url.hostname, parsed_url.port), timeout=5) as s:
        s.sendall(b'\n')


def test_sign_and_send(sgx, account, w3):
    key, address = account
    txn = generate_tx(w3)
    txn['nonce'] = w3.eth.get_transaction_count(address)
    signed_txn = sgx.sign(txn, key)
    tx_hash = w3.eth.send_raw_transaction(signed_txn.raw_transaction)
    w3.eth.wait_for_transaction_receipt(tx_hash)
    assert isinstance(tx_hash, HexBytes)
    assert tx_hash != HexBytes('0x')

    txn_v1 = generate_tx(w3, 1)
    txn_v1['nonce'] = w3.eth.get_transaction_count(address)
    signed_txn = sgx.sign(txn_v1, key)
    tx_hash = w3.eth.send_raw_transaction(signed_txn.raw_transaction)
    w3.eth.wait_for_transaction_receipt(tx_hash)
    assert isinstance(tx_hash, HexBytes)
    assert tx_hash != HexBytes('0x')


def test_sign_and_send_second_type(sgx, account, w3):
    key, address = account
    txn_v2 = generate_tx(w3, 2)
    txn_v2['nonce'] = w3.eth.get_transaction_count(address)
    signed_txn = sgx.sign(txn_v2, key)
    tx_hash = w3.eth.send_raw_transaction(signed_txn.raw_transaction)
    assert isinstance(tx_hash, HexBytes)
    assert tx_hash != HexBytes('0x')


def test_get_info(sgx):
    generated_key = sgx.generate_key()
    assert generated_key.name and generated_key.name[:3] == 'NEK'
    assert generated_key.address and len(generated_key.address) == 42
    assert generated_key.public_key and len(generated_key.public_key) == 130
    key = generated_key.name
    account = sgx.get_account(key)
    assert account.public_key and len(account.public_key) == 130
    assert account.address and len(account.address) == 42


def test_get_server_status(sgx):
    assert sgx.get_server_status() == 0


def test_get_server_version(sgx):
    assert isinstance(sgx.get_server_version(), str)


def test_sign_message(sgx, account, w3):
    key, address = account
    txn = generate_tx(w3, tx_type=None, no_fee=True)

    txn['nonce'] = w3.eth.get_transaction_count(address)
    unsigned_transaction = transactions.serializable_unsigned_transaction_from_dict(txn)
    transaction_hash = unsigned_transaction.hash()

    message = defunct_hash_message(transaction_hash).hex()
    signed_message = sgx.sign_hash(message, key, None)
    assert signed_message.message_hash == HexBytes(message)
    assert len(signed_message.signature) > 2
    assert type(signed_message.signature) is HexBytes

    recover_account = w3.eth.account.recover_message(
        encode_defunct(transaction_hash), signature=signed_message.signature
    )

    assert recover_account == address


def test_import_ecdsa(sgx, w3):
    random_key_name = secrets.token_hex(32)

    ecdsa_key_name = 'NEK:' + random_key_name

    insecure_ecdsa_private_key = 'f253bad7b1f62b8ff60bbf451cf2e8e9ebb5d6e9bff450c55b8d5504b8c63d3'

    public_key = sgx.import_ecdsa_private_key(ecdsa_key_name, insecure_ecdsa_private_key)

    assert len(public_key) > 0

    assert public_key == sgx.sgx_rpc_server.get_public_key(ecdsa_key_name)

    account = sgx.get_account(ecdsa_key_name).address

    send_eth_w3(w3, account, ETH_VALUE_FOR_TESTS)

    txn = generate_tx(w3, tx_type=None)
    txn['nonce'] = w3.eth.get_transaction_count(account)
    unsigned_transaction = transactions.serializable_unsigned_transaction_from_dict(txn)
    transaction_hash = unsigned_transaction.hash()
    message = defunct_hash_message(transaction_hash).hex()

    signed_message = sgx.sign_hash(message, ecdsa_key_name, None)
    assert signed_message.message_hash == HexBytes(message)
    assert len(signed_message.signature) > 2
    assert type(signed_message.signature) is HexBytes

    recover_account = w3.eth.account.recover_message(
        encode_defunct(transaction_hash), signature=signed_message.signature
    )
    assert recover_account == account



from setuptools import (
    find_packages,
    setup,
)

extras_require = {
    'linter': [
        'ruff==0.13.2',
    ],
    'dev': [
        'coincurve==13.0.0',
        'python-dotenv==0.13.0',
        'twine==6.2.0',
        'pytest==8.4.2',
        'mock==5.2.0',
    ],
}

extras_require['dev'] = extras_require['linter'] + extras_require['dev']

setup(
    name='sgx.py',
    version='0.11b0',
    description='SGX',
    url='http://github.com/skalenetwork/sgx.py',
    author='SKALE Labs',
    author_email='support@skalelabs.com',
    install_requires=[
        'web3>=6.20.2,<8.0.0',
        'pyzmq==27.1.0',
        'pem==23.1.0',
        'cryptography>=46.0.1,<47.0.0',
    ],
    packages=find_packages(exclude=['tests']),
    python_requires='>=3.11,<4',
    extras_require=extras_require,
    package_data={'sgx': ['generate.sh']},
)



import mock
import pytest
import requests
import urllib3

from sgx.http import send_request
from sgx.http import SgxUnreachableError

URL = 'http://127.0.0.1:1026'
INVALID_URL = 'http://127.0.0.1:1033'

TEST_MAX_RETRIES = 5
MIN_TIME_DELTA = 15


def test_send_request_failed_sgx_down():
    method = 'POST'
    params = {'data': 'test'}
    with pytest.raises(SgxUnreachableError):
        send_request(INVALID_URL, method, params)


def test_send_request_failed_sgx_up():
    method = 'POST'
    params = {'data': 'test'}

    def post_mock(*args, **kwargs):
        err = requests.exceptions.ConnectionError('Test')
        err.args = (urllib3.exceptions.MaxRetryError(URL, 'test'),)
        raise err

    with mock.patch('requests.post', post_mock):
        with pytest.raises(SgxUnreachableError):
            send_request(URL, method, params)
